"""
.. module:: openxps.metadynamics
   :platform: Linux, Windows, macOS
   :synopsis: Biasing potentials applied to extra degrees of freedom.

.. moduleauthor:: Charlles Abreu <craabreu@gmail.com>

"""

import functools
import typing as t

import numpy as np
import openmm as mm
from openmm import unit as mmunit

from .biasing_potential import BiasingPotential
from .extra_dof import ExtraDOF


class MetadynamicsBias(BiasingPotential):
    r"""
    A Metadynamics potential applied to extra degrees of freedom.

    .. math::

        V({\bf x}) = \sum_{i=1}^{N} h_i \exp\left\{
            -\frac{1}{2} \left(
                {\bf x} - {\bf c}_i
            \right)^T [{\bf D}({\boldsymbol \sigma}_i)]^{-2} \left(
                {\bf x} - {\bf c}_i
            \right)
        \right\}

    where :math:`{\bf x}` is the vector of extra degrees of freedom, :math:`h_i` is
    the height of the Gaussian kernel, :math:`{\bf c}_i` is the center of the kernel,
    :math:`{\boldsymbol \sigma}_i` is the bandwidth vector of the kernel, and
    :math:`{\bf D}(\cdot)` builds a diagonal matrix from a vector.

    The bias potential can be interpolated on a grid using natural splines for 1D, 2D,
    or 3D functions. This is done by providing the number of points in each dimension
    of the grid. If the grid is not provided, the bias potential is evaluated directly
    from the mathematical expression, which gets more expensive as the number of added
    kernels increases.

    Parameters
    ----------
    extra_dofs
        The extra degrees of freedom subject to the bias potential.
    bandwidth
        The bandwidth vector of the Gaussian kernel. Each element must have units of
        the corresponding extra degree of freedom.
    initial_height
        The initial height of the Gaussian kernels. It must have units of molar energy.
    temperature
        The temperature of the system (with units).
    bias_factor
        The bias factor that controls the equilibrium bias potential and the rate of
        convergence. It must be larger than 1.
    grid_sizes
        The number of points in each dimension of the grid used to interpolate the bias
        potential.

    Raises
    ------
    ValueError
        If the bias factor is not larger than 1 or if temperature does not have valid
        units.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        extra_dofs: t.Sequence[ExtraDOF],
        bandwidth: t.Sequence[mmunit.Quantity],
        initial_height: mmunit.Quantity,
        temperature: mmunit.Quantity,
        bias_factor: float,
        grid_sizes: t.Union[t.Sequence[int], None],
    ) -> None:
        if bias_factor <= 1:
            raise ValueError("bias_factor must be a float larger than 1")
        self._bandwidth = [
            quantity.value_in_unit(xdof.unit)
            for quantity, xdof in zip(bandwidth, extra_dofs)
        ]
        self._initial_height, self._kb_delta_t = [
            quantity.value_in_unit(mmunit.kilojoules_per_mole)
            for quantity in (
                initial_height,
                mmunit.MOLAR_GAS_CONSTANT_R * temperature * (bias_factor - 1),
            )
        ]
        if grid_sizes is None:
            summands = []
            distances = []
            for xdof, sigma in zip(extra_dofs, self._bandwidth):
                distance = f"{xdof.name}-{xdof.name}_center"
                if xdof.isPeriodic():
                    length = xdof.bounds.period
                    distance = f"{length / np.pi}*sin({np.pi / length}*({distance}))"
                distances.append(distance)
                summands.append(f"({xdof.name}_dist/{sigma})^2")
            function = f"height*exp(-0.5*({'+'.join(summands)}))"
            for xdof, distance in zip(extra_dofs, distances):
                function += f";\n{xdof.name}_dist={distance}"
        else:
            function = f"bias({','.join(xdof.name for xdof in extra_dofs)})"
        super().__init__(
            extra_dofs,
            function,
            functools.partial(mm.CustomCompoundBondForce, len(extra_dofs)),
        )
        if grid_sizes is None:
            self._bias_grid = None
            self.addPerBondParameter("height")
            for xdof in extra_dofs:
                self.addPerBondParameter(f"{xdof.name}_center")
        else:
            self._bias_grid = SplineGrid(extra_dofs, grid_sizes)
            self.addTabulatedFunction("bias", self._bias_grid)

    def _performAddKernel(
        self, context: mm.Context, center: t.Sequence[float], potential: float
    ) -> None:
        height = self._initial_height * np.exp(-potential / self._kb_delta_t)
        if self._bias_grid is None:
            self.addBond(self._extra_dof_indices, [height, *center])
            context.reinitialize(preserveState=True)
        else:
            exponents = []
            for i, xdof in enumerate(self._extra_dofs):
                distances = self._bias_grid.getGridPoints(i) - center[i]
                if xdof.isPeriodic():
                    length = xdof.bounds.upper - xdof.bounds.lower
                    distances = (length / np.pi) * np.sin((np.pi / length) * distances)
                    distances[-1] = distances[0]
                exponents.append(-0.5 * (distances / self._bandwidth[i]) ** 2)
            kernel = height * np.exp(
                exponents[0]
                if len(self._extra_dofs) == 1
                else functools.reduce(np.add.outer, reversed(exponents))
            )
            self._bias_grid.addBias(kernel)
            self.updateParametersInContext(context)

    def initialize(self, context_extra_dofs: t.Sequence[ExtraDOF]) -> None:
        super().initialize(context_extra_dofs)
        if self._bias_grid is not None:
            self.addBond(self._extra_dof_indices, [])

    def getBias(self) -> np.ndarray:
        """
        Get the bias values on a spline grid.

        Returns
        -------
        np.ndarray
            The bias values on the spline grid.
        """
        return self._bias_grid.getBias()


class SplineGrid(mm.TabulatedFunction):
    """
    A grid used to interpolate a bias potential applied to extra degrees of freedom.

    Parameters
    ----------
    extra_dofs
        The extra degrees of freedom subject to the bias potential.
    grid_sizes
        The number of points in each dimension of the grid used to interpolate the bias
        potential.

    Raises
    ------
    ValueError
        If the length of ``extra_dofs`` is not between 1 and 3, if ``grid_sizes`` has
        a different length than ``extra_dofs``, if any extra degree of freedom is not
        bounded, or if the dimensions of the grid are not all periodic or all
        non-periodic.

    Examples
    --------
    Create a 2D grid to interpolate a bias potential applied to a single extra degree
    of freedom:
    >>> import openxps as xps
    >>> from openmm import unit
    >>> import numpy as np
    >>> mass = 1.0 * unit.dalton
    >>> bounds = xps.bounds.Reflective(0, 1, unit.nanometers)
    >>> extra_dof = xps.ExtraDOF("x", unit.nanometers, mass, bounds)
    >>> grid = xps.metadynamics.SplineGrid([extra_dof], [10])
    """

    def __init__(  # pylint: disable=super-init-not-called
        self, extra_dofs: t.Sequence[ExtraDOF], grid_sizes: t.Sequence[int]
    ) -> None:
        num_dims = len(extra_dofs)
        if not 1 <= num_dims <= 3:
            raise ValueError("Spline grid can only interpolate 1D, 2D, or 3D functions")
        if len(grid_sizes) != num_dims:
            raise ValueError("Number of grid sizes must match number of extra DOFs")
        if any(xdof.bounds is None for xdof in extra_dofs):
            raise ValueError("Spline interpolation requires bounded extra DOFs")

        self._grid_sizes = np.array(grid_sizes)
        self._grid = [
            np.linspace(xdof.bounds.lower, xdof.bounds.upper, num=grid_size)
            for xdof, grid_size in zip(extra_dofs, grid_sizes)
        ]

        num_periodics = sum(xdof.isPeriodic() for xdof in extra_dofs)
        if num_periodics not in [0, num_dims]:
            raise ValueError("Mixed periodic/non-periodic dimensions are not supported")

        self._widths = [] if num_dims == 1 else grid_sizes
        self._bias = np.zeros(grid_sizes)
        self._limits = []
        for xdof in extra_dofs:
            if xdof.bounds is None:
                raise ValueError("Interpolation requires bounded extra DOFs")
            self._limits.extend([xdof.bounds.lower, xdof.bounds.upper])
        periodic = num_periodics == num_dims
        self._table = getattr(mm, f"Continuous{num_dims}DFunction")(
            *self._widths, self._bias.flatten(), *self._limits, periodic
        )
        self.this = self._table.this

    def getGridPoints(self, dimension: int) -> np.ndarray:
        """
        Get the grid points along a specific dimension.

        Parameters
        ----------
        dimension
            The dimension for which the grid points will be returned.

        Returns
        -------
        np.ndarray
            The grid points along the specified dimension.
        """
        return self._grid[dimension]

    def getBias(self) -> np.ndarray:
        """
        Get the bias values on the spline grid.

        Returns
        -------
        np.ndarray
            The bias values on the spline grid.
        """
        return self._bias

    def addBias(self, bias: np.ndarray) -> None:
        """
        Add bias values to the grid.

        Parameters
        ----------
        bias
            The bias values to be added to the grid.
        """
        self._bias += bias
        self._table.setFunctionParameters(
            *self._widths, self._bias.flatten(), *self._limits
        )



"""
.. module:: openxps.bounds
   :platform: Linux, MacOS, Windows
   :synopsis: Specification of boundary conditions for extra degrees of freedom

.. classauthor:: Charlles Abreu <craabreu@gmail.com>

"""

import typing as t
from dataclasses import dataclass

import cvpack
import numpy as np
from cvpack.serialization import Serializable
from cvpack.units import Quantity
from openmm import unit as mmunit

from .utils import preprocess_args


@dataclass(frozen=True, eq=False)
class Bounds(Serializable):
    """
    A boundary condition for an extra degree of freedom.

    Parameters
    ----------
    lower
        The lower bound for the extra degree of freedom.
    upper
        The upper bound for the extra degree of freedom.
    unit
        The unity of measurement of the bounds. If the bounds do not have a unit, use
        ``dimensionless``.
    """

    lower: float
    upper: float
    unit: cvpack.units.Unit

    def __post_init__(self) -> None:
        for kind in ("lower", "upper"):
            if not isinstance(getattr(self, kind), (int, float)):
                raise TypeError(f"The {kind} bound must be a real number.")
        if self.lower >= self.upper:
            raise ValueError("The upper bound must be greater than the lower bound.")
        if not mmunit.is_unit(self.unit):
            raise TypeError("The unit must be a valid OpenMM unit.")
        object.__setattr__(self, "unit", cvpack.units.Unit(self.unit))

    def __getstate__(self) -> t.Dict[str, t.Any]:
        return {"lower": self.lower, "upper": self.upper, "unit": self.unit}

    def __setstate__(self, keywords: t.Dict[str, t.Any]) -> None:
        self.__init__(**keywords)

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, Bounds):
            return False
        return (
            self.lower * self.unit == other.lower * other.unit
            and self.upper * self.unit == other.upper * other.unit
        )

    def convert(self, unit: mmunit.Unit) -> "Bounds":
        """
        Convert the bounds to a different unit of measurement.

        Parameters
        ----------
        unit
            The new unit of measurement for the bounds.

        Returns
        -------
        Bounds
            The bounds in the new unit of measurement.

        Example
        -------
        >>> import openxps as xps
        >>> from openmm import unit
        >>> bounds = xps.bounds.Periodic(-180, 180, unit.degree)
        >>> bounds.convert(unit.radian)
        Periodic(lower=-3.14159..., upper=3.14159..., unit=rad)
        """
        factor = 1.0 * self.unit / unit
        if not isinstance(factor, float):
            raise ValueError("The unit must be compatible with the bounds unit.")
        return type(self)(factor * self.lower, factor * self.upper, unit)

    def asQuantity(self) -> mmunit.Quantity:
        """
        Return the bounds as a Quantity object.

        Returns
        -------
        Quantity
            The bounds as a Quantity object.

        Example
        -------
        >>> import openxps as xps
        >>> from openmm import unit
        >>> bounds = xps.bounds.Periodic(-180, 180, unit.degree)
        >>> bounds.asQuantity()
        (-180, 180) deg
        """
        return Quantity((self.lower, self.upper), self.unit)

    def leptonExpression(self, variable: str) -> str:
        """
        Return a lepton expression representing the transformation from an unwrapped
        variable to a wrapped value under the boundary conditions.

        Parameters
        ----------
        variable
            The name of the variable in the expression.
        Returns
        -------
        str
            A string representing the transformation.

        Example
        -------
        >>> import openxps as xps
        >>> from openmm import unit
        >>> periodic = xps.bounds.Periodic(-180, 180, unit.degree)
        >>> print(periodic.leptonExpression("x"))
        (scaled_x-floor(scaled_x))*360-180;
        scaled_x=(x+180)/360
        >>> reflective = xps.bounds.Reflective(1, 10, unit.angstrom)
        >>> print(reflective.leptonExpression("y"))
        min(wrapped_y,1-wrapped_y)*18+1;
        wrapped_y=scaled_y-floor(scaled_y);
        scaled_y=(y-1)/18
        """
        raise NotImplementedError(
            "The method transformation must be implemented in subclasses."
        )

    def wrap(self, value: float, rate: float) -> t.Tuple[float, float]:
        """
        Wrap a value around the bounds and adjust its rate of change.

        Parameters
        ----------
        value
            The unwrapped value, in the same unit of measurement as the bounds.
        rate
            The rate of change of the unwrapped value, in the same unit of measurement
            as the bounds divided by a time unit.

        Returns
        -------
        float
            The wrapped value, in the same unit of measurement as the bounds.
        float
            The adjusted rate of change of the wrapped value, in the same unit of
            measurement as the original rate.
        """
        raise NotImplementedError("The method wrap must be implemented in subclasses.")


Bounds.__init__ = preprocess_args(Bounds.__init__)


class Periodic(Bounds):
    """
    A periodic boundary condition. The extra degree of freedom is allowed to wrap
    around the upper and lower bounds.

    Parameters
    ----------
    lower
        The lower bound for the extra degree of freedom.
    upper
        The upper bound for the extra degree of freedom.
    unit
        The unity of measurement of the bounds. If the bounds do not have a unit, use
        ``dimensionless``.

    Example
    -------
    >>> import openxps as xps
    >>> import yaml
    >>> from openmm import unit
    >>> bounds = xps.bounds.Periodic(-180, 180, unit.degree)
    >>> print(bounds)
    Periodic(lower=-180, upper=180, unit=deg)
    >>> assert yaml.safe_load(yaml.safe_dump(bounds)) == bounds
    """

    def __post_init__(self) -> None:
        super().__post_init__()
        self.period = self.upper - self.lower

    def leptonExpression(self, variable: str) -> str:
        scaled = f"scaled_{variable}"
        if self.lower == 0:
            shift = deshift = ""
        elif self.lower > 0:
            shift = f"-{self.lower}"
            deshift = f"+{self.lower}"
        else:
            shift = f"+{-self.lower}"
            deshift = f"-{-self.lower}"
        return (
            f"({scaled}-floor({scaled}))*{self.period}{deshift}"
            f";\n{scaled}=({variable}{shift})/{self.period}"
        )

    def wrap(self, value: float, rate: float) -> t.Tuple[float, float]:
        return (value - self.lower) % self.period + self.lower, rate


Periodic.registerTag("!openxps.bounds.Periodic")


class Reflective(Bounds):
    """
    A reflective boundary condition. The extra degree of freedom collides elastically
    with the upper and lower bounds and is reflected back into the range.

    Parameters
    ----------
    lower
        The lower bound for the extra degree of freedom.
    upper
        The upper bound for the extra degree of freedom.
    unit
        The unity of measurement of the bounds. If the bounds do not have a unit, use
        ``dimensionless``.

    Example
    -------
    >>> import openxps as xps
    >>> from openmm import unit
    >>> bounds = xps.bounds.Reflective(1, 10, unit.angstrom)
    >>> bounds == xps.bounds.Reflective(0.1, 1, unit.nanometer)
    True
    >>> print(bounds)
    Reflective(lower=1, upper=10, unit=A)
    """

    def __post_init__(self) -> None:
        super().__post_init__()
        self.period = 2 * (self.upper - self.lower)

    def leptonExpression(self, variable: str) -> str:
        scaled = f"scaled_{variable}"
        wrapped = f"wrapped_{variable}"
        if self.lower == 0:
            shift = deshift = ""
        elif self.lower > 0:
            shift = f"-{self.lower}"
            deshift = f"+{self.lower}"
        else:
            shift = f"+{-self.lower}"
            deshift = f"-{-self.lower}"
        return (
            f"min({wrapped},1-{wrapped})*{self.period}{deshift}"
            f";\n{wrapped}={scaled}-floor({scaled})"
            f";\n{scaled}=({variable}{shift})/{self.period}"
        )

    def wrap(self, value: float, rate: float) -> t.Tuple[float, float]:
        x = (value - self.lower) % self.period
        if x < self.period - x:
            return x + self.lower, rate
        return self.period - x + self.lower, -rate


Reflective.registerTag("!openxps.bounds.Reflective")


CIRCULAR = Periodic(-np.pi, np.pi, mmunit.radians)



"""
Unit tests for the ExtraDOF class.
"""

import cvpack
import pytest
import yaml
from openmm import unit as mmunit

from openxps.bounds import Periodic, Reflective
from openxps.extra_dof import ExtraDOF


def test_extra_dof_initialization():
    """Test successful ExtraDOF initialization."""
    mass = 3 * mmunit.dalton * (mmunit.nanometer / mmunit.radian) ** 2
    bounds = Periodic(-180, 180, mmunit.degree)
    extra_dof = ExtraDOF("phi", mmunit.radian, mass, bounds)

    assert extra_dof.name == "phi"
    assert extra_dof.unit == mmunit.radian
    assert (
        extra_dof.mass.unit == mmunit.dalton * (mmunit.nanometer / mmunit.radian) ** 2
    )
    assert isinstance(extra_dof.bounds, Periodic)


def test_extra_dof_invalid_unit():
    """Test ExtraDOF initialization with an invalid unit."""
    with pytest.raises(ValueError) as excinfo:
        ExtraDOF(
            "phi",
            "not_a_unit",
            3 * mmunit.dalton * (mmunit.nanometer / mmunit.radian) ** 2,
            None,
        )
    assert "The unit must be a valid OpenMM unit." in str(excinfo.value)


def test_extra_dof_not_md_unit_system():
    """Test ExtraDOF initialization with an incompatible unit."""
    with pytest.raises(ValueError) as excinfo:
        ExtraDOF("phi", mmunit.degree, 3 * mmunit.dalton, None)
    assert "Unit degree is incompatible with OpenMM's MD unit system." in str(
        excinfo.value
    )


def test_extra_dof_mass_without_unit():
    """Test ExtraDOF initialization with mass missing a unit."""
    with pytest.raises(TypeError) as excinfo:
        ExtraDOF("phi", mmunit.radian, 3, None)
    assert "Mass must be have units of measurement." in str(excinfo.value)


def test_extra_dof_incompatible_mass_unit():
    """Test ExtraDOF initialization with mass having incompatible units."""
    with pytest.raises(TypeError):
        ExtraDOF("phi", mmunit.radian, 3 * mmunit.meter, None)


def test_extra_dof_bounds_incompatible_units():
    """Test ExtraDOF initialization with bounds having incompatible units."""
    bounds = Reflective(0, 10, mmunit.meter)  # Incompatible with radian
    with pytest.raises(ValueError):
        ExtraDOF(
            "phi",
            mmunit.radian,
            3 * mmunit.dalton * (mmunit.nanometer / mmunit.radian) ** 2,
            bounds,
        )


def test_extra_dof_serialization():
    """Test YAML serialization and deserialization of ExtraDOF."""
    mass = 3 * mmunit.dalton * (mmunit.nanometer / mmunit.radian) ** 2
    bounds = Periodic(-180, 180, mmunit.degree)
    extra_dof = ExtraDOF("phi", mmunit.radian, mass, bounds)

    serialized = yaml.safe_dump(extra_dof)
    deserialized = yaml.safe_load(serialized)

    assert deserialized == extra_dof


def test_extra_dof_without_bounds():
    """Test ExtraDOF initialization without bounds."""
    mass = 3 * mmunit.dalton * (mmunit.nanometer / mmunit.radian) ** 2
    extra_dof = ExtraDOF("phi", mmunit.radian, mass, None)

    assert extra_dof.bounds is None


def test_extra_dof_bounds_type_error():
    """Test ExtraDOF initialization with incorrect bounds type."""
    with pytest.raises(TypeError):
        ExtraDOF(
            "phi",
            mmunit.radian,
            3 * mmunit.dalton * (mmunit.nanometer / mmunit.radian) ** 2,
            "not_bounds",
        )


def test_extra_dof_distance_method():
    """
    Test the distanceTo method of ExtraDOF.
    """
    psi0 = ExtraDOF(
        "psi0",
        mmunit.radian,
        3 * mmunit.dalton * (mmunit.nanometer / mmunit.radian) ** 2,
        Periodic(-180, 180, mmunit.degree),
    )
    phi0 = ExtraDOF(
        "phi0",
        mmunit.radian,
        3 * mmunit.dalton * (mmunit.nanometer / mmunit.radian) ** 2,
        Periodic(-180, 180, mmunit.degree),
    )

    assert psi0.distanceTo(cvpack.Torsion(6, 8, 14, 16, name="psi")) == (
        "(psi-psi0-6.283185307179586*floor(0.5+(psi-psi0)/6.283185307179586))"
    )

    assert (
        psi0.distanceTo(phi0)
        == "(phi0-psi0-6.283185307179586*floor(0.5+(phi0-psi0)/6.283185307179586))"
    )

    with pytest.raises(TypeError) as excinfo:
        psi0.distanceTo("not_cv")
    assert "Method distanceTo not implemented for type" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        psi0.distanceTo(cvpack.Distance(0, 1))
    assert "Incompatible boundary conditions." in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        psi0.distanceTo(ExtraDOF("x", mmunit.nanometer, 3 * mmunit.dalton, None))
    assert "Incompatible boundary conditions." in str(excinfo.value)

    distance0 = ExtraDOF(
        "distance0",
        mmunit.nanometer,
        3 * mmunit.dalton,
        Reflective(0, 1, mmunit.nanometer),
    )
    assert distance0.distanceTo(cvpack.Distance(0, 1)) == "(distance-distance0)"

    distance1 = ExtraDOF("distance1", mmunit.nanometer, 3 * mmunit.dalton, None)
    assert distance1.distanceTo(distance0) == "(distance0-distance1)"



"""
.. module:: openxps.context
   :platform: Linux, MacOS, Windows
   :synopsis: Context for extended phase-space simulations with OpenMM.

.. classauthor:: Charlles Abreu <craabreu@gmail.com>

"""

import typing as t
from copy import copy
from functools import partial
from types import MethodType

import cvpack
import openmm as mm
from openmm import _openmm as mmswig
from openmm import unit as mmunit

from .biasing_potential import BiasingPotential
from .extra_dof import ExtraDOF


class ExtendedSpaceContext(mm.Context):  # pylint: disable=too-many-instance-attributes
    """
    Wraps an :OpenMM:`Context` object to include extra degrees of freedom (DOFs) and
    allow for extended phase-space (XPS) simulations.

    **Note**: The system and integrator attached to the context are modified in-place.

    A provided :CVPack:`MetaCollectiveVariable` is added to the system to couple the
    physical DOFs and the extra ones. The integrator's ``step`` method is replaced with
    a custom function that advances the physical and extension systems in tandem.

    Parameters
    ----------
    context
        The original OpenMM context containing the physical system.
    extra_dofs
        A group of extra degrees of freedom to be included in the XPS simulation.
    coupling_potential
        A meta-collective variable defining the potential energy term that couples the
        physical system to the extra DOFs. It must have units of ``kilojoules_per_mole``
        or equivalent.
    integrator_template
        An :OpenMM:`Integrator` object to be used as a template for the algorithm that
        advances the extra DOFs. If not provided, the physical system's integrator is
        used as a template.
    biasing_potential
        A bias potential applied to the extra DOFs. If not provided, no bias is applied.

    Example
    -------
    >>> import openxps as xps
    >>> from math import pi
    >>> import cvpack
    >>> import openmm
    >>> from openmm import unit
    >>> from openmmtools import testsystems
    >>> model = testsystems.AlanineDipeptideVacuum()
    >>> umbrella_potential = cvpack.MetaCollectiveVariable(
    ...     f"0.5*kappa*min(delta,{2*pi}-delta)^2; delta=abs(phi-phi0)",
    ...     [cvpack.Torsion(6, 8, 14, 16, name="phi")],
    ...     unit.kilojoule_per_mole,
    ...     kappa=1000 * unit.kilojoule_per_mole / unit.radian**2,
    ...     phi0=pi*unit.radian,
    ... )
    >>> temp = 300 * unit.kelvin
    >>> integrator = openmm.LangevinMiddleIntegrator(
    ...     temp, 1 / unit.picosecond, 4 * unit.femtosecond
    ... )
    >>> platform = openmm.Platform.getPlatformByName("Reference")
    >>> mass = 3 * unit.dalton*(unit.nanometer/unit.radian)**2
    >>> phi0 = xps.ExtraDOF("phi0", unit.radian, mass, xps.bounds.CIRCULAR)
    >>> height = 2 * unit.kilojoule_per_mole
    >>> sigma = 18 * unit.degree
    >>> context = xps.ExtendedSpaceContext(
    ...     openmm.Context(model.system, integrator, platform),
    ...     [phi0],
    ...     umbrella_potential,
    ...     biasing_potential=xps.MetadynamicsBias(
    ...         [phi0], [sigma], height, temp, 10, [100]
    ...     ),
    ... )
    >>> context.setPositions(model.positions)
    >>> context.setVelocitiesToTemperature(temp)
    >>> context.setExtraValues([180 * unit.degree])
    >>> context.setExtraVelocitiesToTemperature(temp)
    >>> context.getIntegrator().step(100)
    >>> context.getExtraValues()
    (Quantity(value=..., unit=radian),)
    >>> context.addBiasKernel()
    >>> state = context.getExtensionContext().getState(getEnergy=True)
    >>> state.getPotentialEnergy()
    Quantity(value=..., unit=kilojoule/mole)
    """

    def __init__(  # pylint: disable=super-init-not-called,too-many-arguments
        self,
        context: mm.Context,
        extra_dofs: t.Sequence[ExtraDOF],
        coupling_potential: cvpack.MetaCollectiveVariable,
        integrator_template: t.Optional[mm.Integrator] = None,
        biasing_potential: t.Optional[BiasingPotential] = None,
    ) -> None:
        self.this = context.this
        self._system = context.getSystem()
        self._integrator = context.getIntegrator()
        self._extra_dofs = tuple(extra_dofs)
        self._coupling_potential = coupling_potential
        self._validate()
        self._coupling_potential.addToSystem(self._system)
        self.reinitialize(preserveState=True)
        self._biasing_potential = biasing_potential
        self._extension_context = self._createExtensionContext(integrator_template)

        self._integrator.step = MethodType(
            partial(
                integrate_extended_space,
                extra_dofs=self._extra_dofs,
                extension_context=self._extension_context,
                coupling_potential=coupling_potential,
            ),
            self,
        )

    def _validate(self) -> None:
        if not all(isinstance(xdof, ExtraDOF) for xdof in self._extra_dofs):
            raise TypeError(
                "All extra degrees of freedom must be instances of ExtraDOF."
            )
        if not isinstance(self._coupling_potential, cvpack.MetaCollectiveVariable):
            raise TypeError(
                "The coupling potential must be an instance of MetaCollectiveVariable."
            )
        if not self._coupling_potential.getUnit().is_compatible(
            mmunit.kilojoule_per_mole
        ):
            raise ValueError("The coupling potential must have units of molar energy.")
        context_parameters = set(self.getParameters())
        force_parameters = self._coupling_potential.getParameterDefaultValues()
        parameter_units = {
            name: quantity.unit for name, quantity in force_parameters.items()
        }
        if parameters := sorted(set(parameter_units) & context_parameters):
            raise ValueError(
                f"The context already contains {parameters} among its parameters."
            )
        xdof_units = {xdof.name: xdof.unit for xdof in self._extra_dofs}
        if parameters := sorted(set(xdof_units) - set(parameter_units)):
            raise ValueError(
                f"The coupling potential parameters do not include {parameters}."
            )
        for name, unit in xdof_units.items():
            if not unit.is_compatible(parameter_units[name]):
                raise ValueError(f"Unit mismatch for parameter '{name}'.")

    def _createExtensionContext(
        self, integrator_template: t.Union[mm.Integrator, None]
    ) -> mm.Context:
        extension_integrator = copy(integrator_template or self._integrator)
        extension_integrator.setStepSize(self._integrator.getStepSize())

        extension_system = mm.System()
        for xdof in self._extra_dofs:
            extension_system.addParticle(xdof.mass / xdof.mass.unit)

        meta_cv = self._coupling_potential
        parameters = meta_cv.getParameterDefaultValues()
        for xdof in self._extra_dofs:
            parameters.pop(xdof.name)
        parameters.update(meta_cv.getInnerValues(self))

        flipped_potential = cvpack.MetaCollectiveVariable(
            function=meta_cv.getEnergyFunction(),
            variables=[
                xdof.createCollectiveVariable(index)
                for index, xdof in enumerate(self._extra_dofs)
            ],
            unit=meta_cv.getUnit(),
            periodicBounds=meta_cv.getPeriodicBounds(),
            name=meta_cv.getName(),
            **parameters,
        )
        flipped_potential.addToSystem(extension_system)

        if self._biasing_potential is not None:
            self._biasing_potential.initialize(self._extra_dofs)
            self._biasing_potential.addToSystem(extension_system)

        return mm.Context(
            extension_system,
            extension_integrator,
            mm.Platform.getPlatformByName("Reference"),
        )

    def addBiasKernel(self) -> None:
        """
        Add a Gaussian kernel to the biasing potential.
        """
        try:
            self._biasing_potential.addKernel(self._extension_context)
        except AttributeError as error:
            raise AttributeError(
                "No biasing potential was provided when creating the context."
            ) from error

    def getExtraDOFs(self) -> t.Tuple[ExtraDOF]:
        """
        Get the extra degrees of freedom included in the extended phase-space system.

        Returns
        -------
        t.Tuple[ExtraDOF]
            A tuple containing the extra degrees of freedom.
        """
        return self._extra_dofs

    def setPositions(self, positions: cvpack.units.MatrixQuantity) -> None:
        """
        Sets the positions of all particles in the physical system.

        Parameters
        ----------
        positions
            The positions for each particle in the system.
        """
        super().setPositions(positions)
        for name, value in self._coupling_potential.getInnerValues(self).items():
            self._extension_context.setParameter(name, value / value.unit)

    def setExtraValues(self, values: t.Iterable[mmunit.Quantity]) -> None:
        """
        Set the values of the extra degrees of freedom.

        Parameters
        ----------
        values
            A sequence of quantities containing the values and units of all extra
            degrees of freedom.
        """
        positions = []
        for xdof, value in zip(self._extra_dofs, values):
            if mmunit.is_quantity(value):
                value = value.value_in_unit(xdof.unit)
            positions.append(mm.Vec3(value, 0, 0))
            if xdof.bounds is not None:
                value, _ = xdof.bounds.wrap(value, 0)
            self.setParameter(xdof.name, value)
        self._extension_context.setPositions(positions)
        for name, value in self._coupling_potential.getInnerValues(self).items():
            self._extension_context.setParameter(name, value / value.unit)

    def getExtraValues(self) -> t.Tuple[mmunit.Quantity]:
        """
        Get the values of the extra degrees of freedom.

        Returns
        -------
        t.Tuple[mmunit.Quantity]
            A tuple containing the values of the extra degrees of freedom.
        """
        return tuple(
            self.getParameter(xdof.name) * xdof.unit for xdof in self._extra_dofs
        )

    def setExtraVelocities(self, velocities: t.Iterable[mmunit.Quantity]) -> None:
        """
        Set the velocities of the extra degrees of freedom.

        Parameters
        ----------
        velocities
            A dictionary containing the velocities of the extra degrees of freedom.
        """
        velocities = list(velocities)
        for i, xdof in enumerate(self._extra_dofs):
            value = velocities[i]
            if mmunit.is_quantity(value):
                value = value.value_in_unit(xdof.unit / mmunit.picosecond)
            velocities[i] = mm.Vec3(value, 0, 0)
        self._extension_context.setVelocities(velocities)

    def setExtraVelocitiesToTemperature(
        self, temperature: mmunit.Quantity, seed: t.Optional[int] = None
    ) -> None:
        """
        Set the velocities of the extra degrees of freedom to a temperature.

        Parameters
        ----------
        temperature
            The temperature to set the velocities to.
        """
        args = (temperature,) if seed is None else (temperature, seed)
        self._extension_context.setVelocitiesToTemperature(*args)
        state = mmswig.Context_getState(self._extension_context, mm.State.Velocities)
        velocities = mmswig.State__getVectorAsVec3(state, mm.State.Velocities)
        self._extension_context.setVelocities([mm.Vec3(v.x, 0, 0) for v in velocities])

    def getExtraVelocities(self) -> t.Tuple[mmunit.Quantity]:
        """
        Get the velocities of the extra degrees of freedom.

        Returns
        -------
        t.Tuple[mmunit.Quantity]
            A tuple containing the velocities of the extra degrees of freedom.
        """
        state = mmswig.Context_getState(
            self._extension_context, mm.State.Positions | mm.State.Velocities
        )
        positions = mmswig.State__getVectorAsVec3(state, mm.State.Positions)
        velocities = mmswig.State__getVectorAsVec3(state, mm.State.Velocities)
        for i, xdof in enumerate(self._extra_dofs):
            value = positions[i].x
            rate = velocities[i].x
            if xdof.bounds is not None:
                value, rate = xdof.bounds.wrap(value, rate)
            velocities[i] = rate * xdof.unit / mmunit.picosecond
        return tuple(velocities)

    def getExtensionContext(self) -> mm.Context:
        """
        Get a reference to the OpenMM context containing the extension system.

        Returns
        -------
        mm.Context
            The context containing the extension system.
        """
        return self._extension_context


def integrate_extended_space(
    physical_context: mm.Context,
    steps: int,
    extra_dofs: t.Tuple[ExtraDOF],
    extension_context: mm.Context,
    coupling_potential: cvpack.MetaCollectiveVariable,
) -> None:
    """
    Advances the extended phase-space simulation by integrating the physical and
    extension systems, in tandem, over a specified number of time steps.

    Parameters
    ----------
    physical_context
        The OpenMM context containing the physical system.
    steps
        The number of time steps to advance the simulation.
    extra_dofs
        The extra degrees of freedom included in the extended phase-space system.
    extension_context : mm.Context
        The OpenMM context containing the extension system.
    coupling_potential
        The potential that couples the physical and extra degrees of freedom.

    Raises
    ------
    mm.OpenMMException
        If the particle positions or extra degrees of freedom have not been properly
        initialized in the context.
    """

    for _ in range(steps):
        # pylint: disable=protected-access
        mmswig.Integrator_step(physical_context._integrator, 1)
        mmswig.Integrator_step(extension_context._integrator, 1)
        # pylint: enable=protected-access

        state = mmswig.Context_getState(extension_context, mm.State.Positions)
        positions = mmswig.State__getVectorAsVec3(state, mm.State.Positions)
        for i, xdof in enumerate(extra_dofs):
            value = positions[i].x
            if xdof.bounds is not None:
                value, _ = xdof.bounds.wrap(value, 0)
            mmswig.Context_setParameter(physical_context, xdof.name, value)

        collective_variables = mmswig.CustomCVForce_getCollectiveVariableValues(
            coupling_potential,
            physical_context,
        )
        for i, value in enumerate(collective_variables):
            mmswig.Context_setParameter(
                extension_context,
                mmswig.CustomCVForce_getCollectiveVariableName(coupling_potential, i),
                value,
            )



"""
.. module:: openxps.extra_dof
   :platform: Linux, Windows, macOS
   :synopsis: Extra degrees of freedom for XPS simulations

.. moduleauthor:: Charlles Abreu <craabreu@gmail.com>

"""

import typing as t
from dataclasses import dataclass

import cvpack
import openmm as mm
from cvpack.serialization import Serializable
from cvpack.units import Quantity
from openmm import unit as mmunit

from .bounds import Bounds, Periodic
from .utils import preprocess_args


@dataclass(frozen=True)
class ExtraDOF(Serializable):
    """
    Extra degree of freedom for extended phase-space simulations with OpenMM.

    Parameters
    ----------
    name
        The name of the context parameter to be turned into an extra degree of
        freedom.
    unit
        The unity of measurement of this extra degree of freedom. It must be
        compatible with OpenMM's MD unit system (mass in ``dalton``, distance in
        ``nanometer``, angle in ``radian``, time in ``picosecond``, temperature in
        ``kelvin``, energy in ``kilojoules_per_mol``). If the extra degree of
        freedom does not have a unit, use ``dimensionless``.
    mass
        The mass assigned to this extra degree of freedom, whose unit of measurement
        must be compatible with ``dalton*(nanometer/unit)**2``, where ``unit`` is the
        extra degree of freedom's own unit (see above).
    bounds
        The boundary condition to be applied to this extra degree of freedom. It must
        be an instance of ``openxps.bounds.Periodic``, ``openxps.bounds.Reflective``,
        or ``None`` (for unbounded variables). If it is not ``None``, its unit of
        measurement must be compatible with the extra degree of freedom's own unit.

    Example
    -------
    >>> import openxps as xps
    >>> import yaml
    >>> from openmm import unit
    >>> dv = xps.ExtraDOF(
    ...     "psi",
    ...     unit.radian,
    ...     3 * unit.dalton*(unit.nanometer/unit.radian)**2,
    ...     xps.bounds.Periodic(-180, 180, unit.degree)
    ... )
    >>> dv
    ExtraDOF(name='psi', unit=rad, mass=3 nm**2 Da/(rad**2), bounds=...)
    >>> dv.bounds
    Periodic(lower=-3.14159..., upper=3.14159..., unit=rad)
    >>> assert yaml.safe_load(yaml.safe_dump(dv)) == dv
    """

    name: str
    unit: mmunit.Unit
    mass: mmunit.Quantity
    bounds: t.Union[Bounds, None]

    def __post_init__(self) -> None:
        if not mmunit.is_unit(self.unit):
            raise ValueError("The unit must be a valid OpenMM unit.")
        if (1 * self.unit).value_in_unit_system(mmunit.md_unit_system) != 1:
            raise ValueError(
                f"Unit {self.unit} is incompatible with OpenMM's MD unit system."
            )

        if not mmunit.is_quantity(self.mass):
            raise TypeError("Mass must be have units of measurement.")
        mass_unit = mmunit.dalton * (mmunit.nanometer / self.unit) ** 2
        if not self.mass.unit.is_compatible(mass_unit):
            raise TypeError(f"Mass units must be compatible with {mass_unit}.")
        object.__setattr__(self, "mass", Quantity(self.mass.in_units_of(mass_unit)))

        if isinstance(self.bounds, Bounds):
            if not self.bounds.unit.is_compatible(self.unit):
                raise ValueError("Provided bounds have incompatible units.")
            object.__setattr__(self, "bounds", self.bounds.convert(self.unit))
        elif self.bounds is not None:
            raise TypeError("The bounds must be an instance of Bounds or None.")

    def __getstate__(self) -> t.Dict[str, t.Any]:
        return {
            "name": self.name,
            "unit": self.unit,
            "mass": self.mass,
            "bounds": self.bounds,
        }

    def __setstate__(self, keywords: t.Dict[str, t.Any]) -> None:
        self.__init__(**keywords)

    def isPeriodic(self) -> bool:
        """
        Returns whether this extra degree of freedom is periodic.

        Returns
        -------
        bool
            ``True`` if this extra degree of freedom is periodic, ``False`` otherwise.

        Example
        -------
        >>> import openxps as xps
        >>> from openmm import unit
        >>> dv = xps.ExtraDOF(
        ...     "psi",
        ...     unit.radian,
        ...     3 * unit.dalton*(unit.nanometer/unit.radian)**2,
        ...     xps.bounds.Periodic(-180, 180, unit.degree)
        ... )
        >>> dv.isPeriodic()
        True
        """
        return isinstance(self.bounds, Periodic)

    def createCollectiveVariable(
        self, particle: int, name: t.Optional[str] = None
    ) -> cvpack.OpenMMForceWrapper:
        """
        Returns a :CVPack:`OpenMMForceWrapper` object associating this extra degree of
        freedom with the x coordinate of a particle in an OpenMM system.

        Parameters
        ----------
        particle
            The index of the particle whose x coordinate will be associated with this
            extra degree of freedom.
        name
            The name of the context parameter to be used in the OpenMM system. If
            ``None``, the name of the extra degree of freedom will be used.

        Returns
        -------
        cvpack.OpenMMForceWrapper
            The collective variable object representing this extra degree of freedom.

        Example
        -------
        >>> import openxps as xps
        >>> from openmm import unit
        >>> xdof = xps.ExtraDOF(
        ...     "psi",
        ...     unit.radian,
        ...     3 * unit.dalton*(unit.nanometer/unit.radian)**2,
        ...     xps.bounds.Periodic(-180, 180, unit.degree)
        ... )
        >>> cv = xdof.createCollectiveVariable(0)
        """
        bounds = self.bounds
        if bounds is None:
            force = mm.CustomExternalForce("x")
        else:
            force = mm.CustomExternalForce(bounds.leptonExpression("x"))
            bounds = bounds.asQuantity() if self.isPeriodic() else None
        force.addParticle(particle, [])
        return cvpack.OpenMMForceWrapper(
            force,
            self.unit,
            bounds,
            self.name if name is None else name,
        )

    def _distanceToCV(self, other: cvpack.CollectiveVariable) -> str:
        name = other.getName()
        diff = f"{name}-{self.name}"
        if other.getPeriodicBounds() is None and not self.isPeriodic():
            return f"({diff})"
        if self.isPeriodic() and other.getPeriodicBounds() == self.bounds.asQuantity():
            period = self.bounds.period
            return f"({diff}-{period}*floor(0.5+({diff})/{period}))"
        raise ValueError("Incompatible boundary conditions.")

    def _distanceToExtraDOF(self, other: "ExtraDOF") -> str:
        diff = f"{other.name}-{self.name}"
        if not (self.isPeriodic() or other.isPeriodic()):
            return f"({diff})"
        if other.isPeriodic() and self.isPeriodic():
            period = self.bounds.period
            return f"({diff}-{period}*floor(0.5+({diff})/{period}))"
        raise ValueError("Incompatible boundary conditions.")

    def distanceTo(self, other: cvpack.CollectiveVariable) -> str:
        """
        Returns a Lepton expression representing the distance between this extra degree
        of freedom and another variable.

        Parameters
        ----------
        other
            The other variable to which the distance will be calculated.

        Returns
        -------
        str
            A string representing the distance between the two variables.

        Example
        -------
        >>> import openxps as xps
        >>> import pytest
        >>> from openmm import unit
        >>> xdof = xps.ExtraDOF(
        ...     "psi0",
        ...     unit.radian,
        ...     3 * unit.dalton*(unit.nanometer/unit.radian)**2,
        ...     xps.bounds.Periodic(-180, 180, unit.degree)
        ... )
        >>> psi = cvpack.Torsion(6, 8, 14, 16, name="psi")
        >>> xdof.distanceTo(psi)
        '(psi-psi0-6.28318...*floor(0.5+(psi-psi0)/6.28318...))'
        >>>

        """
        if isinstance(other, cvpack.CollectiveVariable):
            return self._distanceToCV(other)
        if isinstance(other, ExtraDOF):
            return self._distanceToExtraDOF(other)
        raise TypeError(f"Method distanceTo not implemented for type {type(other)}.")


ExtraDOF.__init__ = preprocess_args(ExtraDOF.__init__)

ExtraDOF.registerTag("!openxps.ExtraDOF")



"""
.. module:: openxps.state_data_writer
   :platform: Linux, MacOS, Windows
   :synopsis: A custom writer for reporting state data from an external context.

.. classauthor:: Charlles Abreu <craabreu@gmail.com>

"""

import typing as t

import openmm as mm
from cvpack.reporting.custom_writer import CustomWriter
from openmm import _openmm as mmswig
from openmm import app as mmapp
from openmm import unit as mmunit

from .context import ExtendedSpaceContext


class ExtensionWriter(CustomWriter):  # pylint: disable=too-many-instance-attributes
    """
    A custom writer for reporting state data from an extension context.

    Parameters
    ----------
    context
        The extended space context whose extension-context state data will be reported.
    potential
        If ``True``, the potential energy of the extension context will be reported.
    kinetic
        If ``True``, the kinetic energy of the extension context will be reported.
    total
        If ``True``, the total energy of the extension context will be reported.
    temperature
        If ``True``, the temperature of the extension context will be reported.

    Example
    -------
    >>> import openxps as xps
    >>> from math import pi
    >>> from sys import stdout
    >>> import cvpack
    >>> import openmm
    >>> from openmm import app, unit
    >>> from openmmtools import testsystems
    >>> model = testsystems.AlanineDipeptideVacuum()
    >>> umbrella_potential = cvpack.MetaCollectiveVariable(
    ...     f"0.5*kappa*min(delta,{2*pi}-delta)^2; delta=abs(phi-phi0)",
    ...     [cvpack.Torsion(6, 8, 14, 16, name="phi")],
    ...     unit.kilojoule_per_mole,
    ...     kappa=1000 * unit.kilojoule_per_mole / unit.radian**2,
    ...     phi0=pi*unit.radian,
    ... )
    >>> integrator = openmm.LangevinMiddleIntegrator(
    ...     300 * unit.kelvin, 1 / unit.picosecond, 4 * unit.femtosecond
    ... )
    >>> integrator.setRandomNumberSeed(1234)
    >>> platform = openmm.Platform.getPlatformByName("Reference")
    >>> simulation = app.Simulation(
    ...     model.topology, model.system, integrator, platform
    ... )
    >>> mass = 3 * unit.dalton*(unit.nanometer/unit.radian)**2
    >>> phi0 = xps.ExtraDOF("phi0", unit.radian, mass, xps.bounds.CIRCULAR)
    >>> context = xps.ExtendedSpaceContext(
    ...     simulation.context, [phi0], umbrella_potential
    ... )
    >>> context.setPositions(model.positions)
    >>> context.setVelocitiesToTemperature(300 * unit.kelvin, 1234)
    >>> context.setExtraValues([180 * unit.degree])
    >>> context.setExtraVelocitiesToTemperature(300 * unit.kelvin, 1234)
    >>> reporter = cvpack.reporting.StateDataReporter(
    ...     stdout,
    ...     10,
    ...     step=True,
    ...     kineticEnergy=True,
    ...     writers=[xps.ExtensionWriter(context, kinetic=True)],
    ... )
    >>> simulation.reporters.append(reporter)
    >>> simulation.step(100)  # doctest: +SKIP
    #"Step","Kinetic Energy (kJ/mole)","Extension Kinetic Energy (kJ/mole)"
    10,60.512...,1.7013...
    20,75.765...,2.5089...
    30,61.116...,1.3375...
    40,52.359...,0.4791...
    50,61.382...,0.7065...
    60,48.674...,0.6520...
    70,60.771...,1.3525...
    80,46.518...,2.0280...
    90,66.111...,0.9597...
    100,60.94...,0.9695...
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        context: ExtendedSpaceContext,
        potential: bool = False,
        kinetic: bool = False,
        total: bool = False,
        temperature: bool = False,
    ) -> None:
        if not isinstance(context, ExtendedSpaceContext):
            raise TypeError("The context must be an instance of ExtendedSpaceContext.")
        self._context = context
        self._potential = potential
        self._kinetic = kinetic
        self._total = total
        self._temperature = temperature
        self._needs_energy = potential or kinetic or total or temperature
        self._needs_velocities = kinetic or temperature
        self._temp_factor = 0.0

    def initialize(self, simulation: mmapp.Simulation) -> None:
        if self._temperature:
            number = len(self._context.getExtraDOFs())
            kb = mmunit.MOLAR_GAS_CONSTANT_R.value_in_unit(
                mmunit.kilojoules_per_mole / mmunit.kelvin
            )
            self._temp_factor = 2 / (number * kb)

    def getHeaders(self) -> t.List[str]:
        headers = []
        if self._potential:
            headers.append("Extension Potential Energy (kJ/mole)")
        if self._kinetic:
            headers.append("Extension Kinetic Energy (kJ/mole)")
        if self._total:
            headers.append("Extension Total Energy (kJ/mole)")
        if self._temperature:
            headers.append("Extension Temperature (K)")
        return headers

    def getValues(self, simulation: mmapp.Simulation) -> t.List[float]:
        state = self._context.getExtensionContext().getState(
            getEnergy=self._needs_energy, getVelocities=self._needs_velocities
        )
        if self._needs_energy:
            potential_energy = mmswig.State_getPotentialEnergy(state)
            kinetic_energy = mmswig.State_getKineticEnergy(state)
        if self._needs_velocities:
            velocities = mmswig.State__getVectorAsVec3(state, mm.State.Velocities)
            for xdof, velocity in zip(self._context.getExtraDOFs(), velocities):
                mass = xdof.mass._value  # pylint: disable=protected-access
                kinetic_energy -= 0.5 * mass * (velocity.y**2 + velocity.z**2)
        values = []
        if self._potential:
            values.append(potential_energy)
        if self._kinetic:
            values.append(kinetic_energy)
        if self._total:
            values.append(potential_energy + kinetic_energy)
        if self._temperature:
            values.append(self._temp_factor * kinetic_energy)
        return values



"""
Unit and regression test for the openxps package.

"""

# Import package, test suite, and other packages as needed
import sys


def test_openxps_imported():
    """Sample test, will always pass so long as import statement worked"""
    assert "openxps" in sys.modules



"""
Extended Phase-Space Methods with OpenMM
"""

from ._version import __version__  # noqa: F401
from .context import ExtendedSpaceContext  # noqa: F401
from .extension_writer import ExtensionWriter  # noqa: F401
from .extra_dof import ExtraDOF  # noqa: F401
from .metadynamics import MetadynamicsBias  # noqa: F401

__all__ = [
    "ExtendedSpaceContext",
    "ExtensionWriter",
    "ExtraDOF",
    "MetadynamicsBias",
]



"""
Test the extension writer
"""

import os
import tempfile
from math import pi

import cvpack
import openmm
import pytest
from openmm import app, unit
from openmmtools import testsystems

import openxps as xps


def test_extension_writer():
    """
    Test the extension writer
    """

    model = testsystems.AlanineDipeptideVacuum()
    umbrella_potential = cvpack.MetaCollectiveVariable(
        f"0.5*kappa*min(delta,{2*pi}-delta)^2; delta=abs(phi-phi0)",
        [cvpack.Torsion(6, 8, 14, 16, name="phi")],
        unit.kilojoule_per_mole,
        kappa=1000 * unit.kilojoule_per_mole / unit.radian**2,
        phi0=pi * unit.radian,
    )
    integrator = openmm.LangevinMiddleIntegrator(
        300 * unit.kelvin, 1 / unit.picosecond, 4 * unit.femtosecond
    )
    integrator.setRandomNumberSeed(1234)
    platform = openmm.Platform.getPlatformByName("Reference")
    simulation = app.Simulation(model.topology, model.system, integrator, platform)
    mass = 3 * unit.dalton * (unit.nanometer / unit.radian) ** 2
    phi0 = xps.ExtraDOF("phi0", unit.radian, mass, xps.bounds.CIRCULAR)
    context = xps.ExtendedSpaceContext(simulation.context, [phi0], umbrella_potential)
    context.setPositions(model.positions)
    context.setVelocitiesToTemperature(300 * unit.kelvin, 1234)
    context.setExtraValues([180 * unit.degree])
    context.setExtraVelocitiesToTemperature(300 * unit.kelvin, 1234)
    with tempfile.TemporaryDirectory() as dirpath:
        with open(os.path.join(dirpath, "report.csv"), "w", encoding="utf-8") as file:
            reporter = cvpack.reporting.StateDataReporter(
                file,
                10,
                step=True,
                writers=[
                    xps.ExtensionWriter(
                        context,
                        potential=True,
                        kinetic=True,
                        total=True,
                        temperature=True,
                    )
                ],
            )
            simulation.reporters.append(reporter)
            simulation.step(100)
        with open(os.path.join(dirpath, "report.csv"), "r", encoding="utf-8") as file:
            assert file.readline() == ",".join(
                [
                    '#"Step"',
                    '"Extension Potential Energy (kJ/mole)"',
                    '"Extension Kinetic Energy (kJ/mole)"',
                    '"Extension Total Energy (kJ/mole)"',
                    '"Extension Temperature (K)"\n',
                ]
            )


def test_raise_exception():
    """
    Test the extension writer raises an exception
    """

    model = testsystems.AlanineDipeptideVacuum()
    integrator = openmm.VerletIntegrator(0.001)
    platform = openmm.Platform.getPlatformByName("Reference")
    context = openmm.Context(model.system, integrator, platform)
    with pytest.raises(TypeError) as excinfo:
        xps.ExtensionWriter(context, potential=True)
    assert "The context must be an instance of ExtendedSpaceContext" in str(
        excinfo.value
    )



"""
Unit tests for the openxps.bounds module.
"""

import numpy as np
import pytest
import yaml
from openmm import unit as mmunit

from openxps.bounds import CIRCULAR, Bounds, Periodic, Reflective


def test_bounds_initialization():
    """
    Test the initialization of Bounds.
    """
    bounds = Bounds(-1, 1, mmunit.meters)
    assert bounds.lower == -1
    assert bounds.upper == 1
    assert bounds.unit == mmunit.meters

    with pytest.raises(ValueError):
        Bounds(1, -1, mmunit.meters)

    with pytest.raises(TypeError):
        Bounds(-1, 1, "not_a_unit")

    with pytest.raises(TypeError):
        Bounds("low", "high", mmunit.meters)


def test_bounds_equality():
    """
    Test the equality of Bounds objects.
    """
    assert Bounds(-1, 1, mmunit.meters) != "not_a_bounds"

    bounds1 = Bounds(-1, 1, mmunit.meters)
    bounds2 = Bounds(-1000, 1000, mmunit.millimeters)
    assert bounds1 == bounds2


def test_bounds_conversion():
    """
    Test the conversion of Bounds units.
    """
    bounds = Bounds(-1, 1, mmunit.meters)
    converted = bounds.convert(mmunit.centimeters)
    assert converted.lower == -100
    assert converted.upper == 100
    assert converted.unit == mmunit.centimeters

    with pytest.raises(ValueError):
        bounds.convert(mmunit.seconds)


def test_periodic_wrap():
    """
    Test the wrapping behavior of Periodic bounds.
    """
    bounds = Periodic(-180, 180, mmunit.degrees)
    wrapped_value, wrapped_rate = bounds.wrap(190, 10)
    assert wrapped_value == -170
    assert wrapped_rate == 10


def test_reflective_wrap():
    """
    Test the wrapping behavior of Reflective bounds.
    """
    bounds = Reflective(0, 10, mmunit.dimensionless)
    wrapped_value, wrapped_rate = bounds.wrap(12, 1)
    assert wrapped_value == 8
    assert wrapped_rate == -1

    wrapped_value, wrapped_rate = bounds.wrap(-2, -1)
    assert wrapped_value == 2
    assert wrapped_rate == 1

    wrapped_value, wrapped_rate = bounds.wrap(20, 1)
    assert wrapped_value == 0
    assert wrapped_rate == 1

    wrapped_value, wrapped_rate = bounds.wrap(0, -1)
    assert wrapped_value == 0
    assert wrapped_rate == -1


def test_circular_constant():
    """
    Test the CIRCULAR constant.
    """
    assert CIRCULAR.lower == -np.pi
    assert CIRCULAR.upper == np.pi
    assert CIRCULAR.unit == mmunit.radians


def test_serialization():
    """
    Test the serialization and deserialization of Bounds objects using YAML.
    """

    periodic = Periodic(-np.pi, np.pi, mmunit.radians)
    serialized = yaml.safe_dump(periodic)
    deserialized = yaml.safe_load(serialized)
    assert deserialized == periodic

    reflective = Reflective(0, 10, mmunit.dimensionless)
    serialized = yaml.safe_dump(reflective)
    deserialized = yaml.safe_load(serialized)
    assert deserialized == reflective


def test_bounds_wrap_not_implemented():
    """
    Test the NotImplementedError when calling the wrap method of Bounds.
    """
    bounds = Bounds(-1, 1, mmunit.meters)
    with pytest.raises(NotImplementedError):
        bounds.wrap(0, 0)


def test_bounds_transformation_not_implemented():
    """
    Test the NotImplementedError when calling the transformation method of Bounds.
    """
    bounds = Bounds(-1, 1, mmunit.meters)
    with pytest.raises(NotImplementedError):
        bounds.leptonExpression("x")



"""
.. module:: openxps.biasing_potential
   :platform: Linux, Windows, macOS
   :synopsis: Abstract class for bias potentials applied to extra degrees of freedom.

.. moduleauthor:: Charlles Abreu <craabreu@gmail.com>

"""

import typing as t

import cvpack
import openmm as mm
from openmm import _openmm as mmswig
from openmm import unit as mmunit

from .extra_dof import ExtraDOF


class BiasingPotential(cvpack.OpenMMForceWrapper):
    """
    Abstract class for bias potentials applied to extra degrees of freedom.

    Parameters
    ----------
    extra_dofs
        The extra degrees of freedom subject to the bias potential.
    function
        The mathematical expression defining the bias potential.
    force_constructor
        A function that constructs the OpenMM force object from the mathematical
        expression.
    """

    def __init__(
        self,
        extra_dofs: t.Sequence[ExtraDOF],
        function: str,
        force_constructor: t.Callable[[str], mm.Force],
    ) -> None:
        self._extra_dofs = tuple(extra_dofs)
        self._extra_dof_indices = self._get_state_args = None
        expression = function
        for index, xdof in enumerate(extra_dofs, start=1):
            variable = f"x{index}"
            if xdof.bounds is not None:
                variable = xdof.bounds.leptonExpression(variable)
            expression += f";{xdof.name}={variable}"
        super().__init__(
            force_constructor(expression),
            mmunit.kilojoules_per_mole,
            name="biasing_potential",
        )

    def _performAddKernel(
        self, context: mm.Context, center: t.Sequence[float], potential: float
    ) -> None:
        raise NotImplementedError(
            "Method _performAddKernel must be implemented by derived classes"
        )

    def initialize(self, context_extra_dofs: t.Sequence[ExtraDOF]) -> None:
        """
        Initialize the bias potential for a specific context.

        Parameters
        ----------
        context_extra_dofs
            All the extra degrees of freedom in the extension context.
        """
        try:
            self._extra_dof_indices = tuple(
                map(context_extra_dofs.index, self._extra_dofs)
            )
        except ValueError as error:
            raise ValueError(
                "Not all extra DOFs in the bias potential are present in the context"
            ) from error
        self._get_state_args = (
            mm.State.Positions | mm.State.Energy,
            False,
            1 << self.getForceGroup(),
        )

    def addKernel(self, context: mm.Context) -> None:
        """
        Add a Gaussian kernel to the bias potential.

        Parameters
        ----------
        context
            The OpenMM context object where the bias potential is applied.
        """
        if self._extra_dof_indices is None:
            raise ValueError("Bias potential has not been initialized")

        state = mmswig.Context_getState(context, *self._get_state_args)
        positions = mmswig.State__getVectorAsVec3(state, mm.State.Positions)
        center = []
        for xdof, index in zip(self._extra_dofs, self._extra_dof_indices):
            value = positions[index].x
            if xdof.bounds is not None:
                value, _ = xdof.bounds.wrap(value, 0)
            center.append(value)
        self._performAddKernel(context, center, mmswig.State_getPotentialEnergy(state))



"""
Unit tests for the openxps.context module.
"""

import cvpack
import numpy as np
import openmm as mm
import pytest
from openmm import unit as mmunit
from openmmtools import testsystems

from openxps.bounds import CIRCULAR, Reflective
from openxps.context import ExtendedSpaceContext
from openxps.extra_dof import ExtraDOF


def create_basic_context(model):
    """Helper function to create a basic OpenMM Context object."""
    integrator = mm.VerletIntegrator(1.0 * mmunit.femtosecond)
    platform = mm.Platform.getPlatformByName("Reference")
    return mm.Context(model.system, integrator, platform)


def create_extra_dofs():
    """Helper function to create an ExtraDOF object."""
    mass = 3 * mmunit.dalton * (mmunit.nanometer / mmunit.radian) ** 2
    kwargs = {"unit": mmunit.nanometer, "mass": 1 * mmunit.dalton}
    return [
        ExtraDOF(name="phi0", unit=mmunit.radian, mass=mass, bounds=CIRCULAR),
        ExtraDOF(name="x0", bounds=None, **kwargs),
        ExtraDOF(name="y0", bounds=Reflective(-1, 1, mmunit.nanometer), **kwargs),
    ]


def create_coupling_potential(
    phi0=180 * mmunit.degrees, unit=mmunit.kilojoule_per_mole
):
    """Helper function to create a MetaCollectiveVariable object."""
    kwargs = {
        "kappa": 1000 * mmunit.kilojoule_per_mole / mmunit.radians**2,
        "alpha": 0.01 * mmunit.kilojoule_per_mole / mmunit.nanometer**2,
        "x0": 1 * mmunit.nanometer,
        "y0": 1 * mmunit.nanometer,
    }
    if phi0 is not None:
        kwargs["phi0"] = phi0
    return cvpack.MetaCollectiveVariable(
        f"0.5*kappa*min(delta_phi,{2*np.pi}-delta_phi)^2+alpha*(x0-y0)^2"
        "; delta_phi=abs(phi-phi0)",
        [cvpack.Torsion(6, 8, 14, 16, name="phi")],
        unit,
        **kwargs,
    )


def create_extended_context(model, coupling_potential=None):
    """Helper function to create an ExtendedSpaceContext object."""
    return ExtendedSpaceContext(
        create_basic_context(model),
        create_extra_dofs(),
        coupling_potential or create_coupling_potential(),
    )


def test_initialization():
    """Test the initialization of ExtendedSpaceContext with basic setup."""
    assert create_extended_context(testsystems.AlanineDipeptideVacuum()) is not None


def test_set_positions_and_velocities():
    """Test setting positions and velocities including extra DOFs."""
    model = testsystems.AlanineDipeptideVacuum()
    context = create_extended_context(model)

    random = np.random.RandomState()  # pylint: disable=no-member
    num_atoms = context.getSystem().getNumParticles()
    positions = model.positions.value_in_unit(mmunit.nanometer)
    velocities = random.uniform(-1, 1, (num_atoms, 3))

    urad = mmunit.radians
    unm = mmunit.nanometers
    ups = mmunit.picoseconds

    context.setPositions(positions)
    context.setVelocities(velocities)
    context.setExtraValues([1 * urad, 0.1 * unm, 0.1 * unm])
    context.setExtraVelocities([1 * urad / ups, 1 * unm / ups, 1 * unm / ups])

    state = context.getState(  # pylint: disable=unexpected-keyword-arg
        getPositions=True, getVelocities=True
    )
    assert state.getPositions(asNumpy=True) == pytest.approx(positions) * unm
    assert state.getVelocities() == pytest.approx(velocities) * unm / ups
    assert context.getExtraValues() == (
        pytest.approx(1) * urad,
        pytest.approx(0.1) * unm,
        pytest.approx(0.1) * unm,
    )
    assert context.getExtraVelocities() == (
        pytest.approx(1) * urad / ups,
        pytest.approx(1) * unm / ups,
        pytest.approx(1) * unm / ups,
    )


def test_raise_exceptions():
    """Test raising exceptions for invalid operations."""
    model = testsystems.AlanineDipeptideVacuum()
    context = create_extended_context(model)

    # with pytest.raises(mm.OpenMMException) as e:
    #     context.setVelocitiesToTemperature(300 * mmunit.kelvin)
    # assert "Particle positions have not been set" in str(e.value)

    context.setPositions(model.positions)
    context.setVelocitiesToTemperature(300 * mmunit.kelvin)

    with pytest.raises(mm.OpenMMException) as e:
        context.setExtraVelocitiesToTemperature(300 * mmunit.kelvin)
    assert "Particle positions have not been set" in str(e.value)

    with pytest.raises(mm.OpenMMException) as e:
        context.getIntegrator().step(1)
    assert "Particle positions have not been set" in str(e.value)

    context.setExtraValues(
        [1 * mmunit.radian, 0.1 * mmunit.nanometer, 0.1 * mmunit.nanometer]
    )
    context.setExtraVelocitiesToTemperature(300 * mmunit.kelvin)

    state = context.getState(  # pylint: disable=unexpected-keyword-arg
        getVelocities=True
    )
    velocities = state.getVelocities()
    extra_velocities = context.getExtraVelocities()
    assert len(velocities) == len(model.positions)
    assert len(extra_velocities) == 3


def test_validation():
    """Test the validation of extended space context."""
    model = testsystems.AlanineDipeptideVacuum()
    context = create_basic_context(model)
    extra_dofs = create_extra_dofs()
    coupling_potential = create_coupling_potential()

    with pytest.raises(TypeError) as e:
        ExtendedSpaceContext(context, [None], coupling_potential)
    assert "extra degrees of freedom must be instances of ExtraDOF" in str(e.value)

    with pytest.raises(TypeError) as e:
        ExtendedSpaceContext(context, extra_dofs, None)
    assert "must be an instance of MetaCollectiveVariable" in str(e.value)

    with pytest.raises(ValueError) as e:
        ExtendedSpaceContext(context, extra_dofs, create_coupling_potential(phi0=None))
    assert "The coupling potential parameters do not include ['phi0']." in str(e.value)

    with pytest.raises(ValueError) as e:
        ExtendedSpaceContext(
            context, extra_dofs, create_coupling_potential(phi0=1 * mmunit.kelvin)
        )
    assert "Unit mismatch for parameter 'phi0'." in str(e.value)

    with pytest.raises(ValueError) as e:
        ExtendedSpaceContext(
            context, extra_dofs, create_coupling_potential(unit=mmunit.radian)
        )
    assert "The coupling potential must have units of molar energy." in str(e.value)

    with pytest.raises(ValueError) as e:
        force = mm.CustomExternalForce("phi0*x")
        force.addGlobalParameter("phi0", 180 * mmunit.degrees)
        force.addGlobalParameter("x0", 1 * mmunit.nanometer)
        force.addGlobalParameter("y0", 1 * mmunit.nanometer)
        context.getSystem().addForce(force)
        context.reinitialize()
        ExtendedSpaceContext(context, extra_dofs, coupling_potential)
    assert (
        "The context already contains ['phi0', 'x0', 'y0'] among its parameters."
        in str(e.value)
    )


def test_consistency():
    """Test the consistency of the extended space context."""
    model = testsystems.AlanineDipeptideVacuum()
    coupling = create_coupling_potential()
    context = create_extended_context(model, coupling)
    context.setPositions(model.positions)
    context.setVelocitiesToTemperature(300 * mmunit.kelvin)
    context.setExtraValues(
        [1000 * mmunit.degrees, 1 * mmunit.nanometer, 1 * mmunit.nanometer]
    )
    context.setExtraVelocitiesToTemperature(300 * mmunit.kelvin)

    for _ in range(10):
        context.getIntegrator().step(1000)

        # pylint: disable=unexpected-keyword-arg
        extension_state = context.getExtensionContext().getState(
            getEnergy=True, getPositions=True, getForces=True
        )
        # pylint: enable=unexpected-keyword-arg

        # Check the consistency of the potential energy
        x1 = extension_state.getPotentialEnergy() / mmunit.kilojoule_per_mole
        x2 = coupling.getValue(context) / mmunit.kilojoule_per_mole
        assert x1 == pytest.approx(x2)

        # Check the consistency of the energy parameter derivatives
        positions = extension_state.getPositions()
        forces = extension_state.getForces()
        x1 = {}
        for i, xdof in enumerate(context.getExtraDOFs()):
            force = forces[i].x
            if xdof.bounds is not None:
                _, force = xdof.bounds.wrap(positions[i].x, force)
            x1[xdof.name] = -force
        x2 = coupling.getParameterDerivatives(context)
        for xdof in context.getExtraDOFs():
            assert x1[xdof.name] == pytest.approx(x2[xdof.name] / x2[xdof.name].unit)



"""
.. module:: openxps.serializable
   :platform: Linux, MacOS, Windows
   :synopsis:

.. classauthor:: Charlles Abreu <craabreu@gmail.com>

"""

import functools
import inspect
import typing as t

from cvpack.units import Quantity, Unit
from openmm import unit as mmunit


def preprocess_args(func: t.Callable) -> t.Callable:
    """
    A decorator that converts instances of unserializable classes to their
    serializable counterparts.

    Parameters
    ----------
        func
            The function to be decorated.

    Returns
    -------
        The decorated function.

    Example
    -------
    >>> from openxps.utils import preprocess_args
    >>> from cvpack import units
    >>> from openmm import unit as mmunit
    >>> @preprocess_args
    ... def function(data):
    ...     return data
    >>> assert isinstance(function(mmunit.angstrom), units.Unit)
    >>> assert isinstance(function(5 * mmunit.angstrom), units.Quantity)
    >>> seq = [mmunit.angstrom, mmunit.nanometer]
    >>> assert isinstance(function(seq), list)
    >>> assert all(isinstance(item, units.Unit) for item in function(seq))
    >>> dct = {"length": 3 * mmunit.angstrom, "time": 2 * mmunit.picosecond}
    >>> assert isinstance(function(dct), dict)
    >>> assert all(isinstance(item, units.Quantity) for item in function(dct).values())
    """
    signature = inspect.signature(func)

    def convert(data: t.Any) -> t.Any:
        if isinstance(data, mmunit.Quantity):
            return Quantity(data)
        if isinstance(data, mmunit.Unit):
            return Unit(data)
        if isinstance(data, t.Sequence) and not isinstance(data, str):
            return type(data)(map(convert, data))
        if isinstance(data, t.Dict):
            return type(data)((key, convert(value)) for key, value in data.items())
        return data

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        bound = signature.bind(*args, **kwargs)
        for name, data in bound.arguments.items():
            bound.arguments[name] = convert(data)
        return func(*bound.args, **bound.kwargs)

    return wrapper
